/***************************************************************************
*
* Copyright (c) 1989, 1990, 1991, 1992, 1993
* by Helsinki University of Technology, Digital Systems Laboratory.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  graph.h
*
* Part of: PROD's programs "gphgen", probe, and strong
*
* Authors: Marko Rauhamaa and Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     .
*
* Version history:
*     $Log$
***************************************************************************/

#define HASH PTRBL

typedef struct GRAPH
{
    char *name;   /* name of the graph */
    BINARYFILE nodes;  /* contains node records and graph globals */
    BINARYFILE arrows; /* contains arrow records */
    BINARYFILE addrs;  /* contains node address table (address of nth node is
                        * in nth position of the table) and table for
                        * strongly connected components (address of 1st node
                        * of nth component is in nth position of the table) */
    unsigned long *glob; /* points to graph globals table (in memory) */
    HASH *hash;          /* points to hash table (in memory) */
#ifdef GPHQUOTA
    BOOL OK; /* cleared when the combined size of the files of the graph
              * exceeds the limit defined in the file GPHQUOTA */
#endif
#ifdef UNIX
#ifdef MAPTOMEM
    int reason;
    MMFILE *mmnodes;
    MMFILE *mmarrows;
    MMFILE *mmaddrs;
#endif
#endif
    unsigned long latest_node; /* address of the latest node asked by Probe */
    unsigned long *latest_rec; /* points to record (in memory) of that node
				* (used for other purpose when graph is
				* constructed) */
    DATATYPE *latest_contents; /* points to contents (in memory) of that node
			     * (alternatively, to the latest arrow contents
			     * asked by Probe) (used for other purpose
			     * when graph is constructed) */
    unsigned long *arrow_rec; /* points to an arrow record in memory */
    unsigned long *sgc_rec; /* points to a strongly connected component
			     * record in memory */
    LIST *list;         /* used when computing strongly connected components */
    unsigned long now;  /* used when computing strongly connected components */
    unsigned long m;    /* used when computing strongly connected components */
    unsigned long *p;   /* used when computing strongly connected components */
    unsigned long *q;   /* used when computing strongly connected components */
    unsigned sgc_val_pos;      /* node record field position */
    unsigned left_child_pos;   /* node record field position */
    unsigned right_child_pos;  /* node record field position */
    unsigned datalen_pos;      /* node record field position */
    unsigned num_node_fields;  /* node record field count */
} GRAPH;

extern TOOLS hleafTools;
              /* = { (CFUNC) NULL, Deallocate, NULL, NULL } */

extern void HLeafInit(/* hleaf */);
/*
    unsigned long **hleaf;
*/

extern void HLeafClear(/* hleaf */);
/*
    unsigned long *hleaf;
*/

extern void HLeafRead(/* hleaf */);
/*
    unsigned long *hleaf;
*/

extern void HLeafWrite(/* hleaf */);
/*
    unsigned long *hleaf;
*/

/* PrimeBelow computes the greatest prime number
 * less than or equal to argument b. */
extern unsigned long PrimeBelow(/* b */);
/*
    unsigned long b;
*/

/* Hash computes hash value for given string. Range is 0L..(prime - 1L). */
extern unsigned long Hash(/* s, prime, coeff */);
/*
    DATATYPE *s;
    unsigned long prime;
    register unsigned long coeff;
*/

/* GphCreate creates a graph. Corresponding files are opened.
 * (Existing files lose their former contents.) Space is reserved
 * for graph globals and the hash table. Hash table size will be
 * the greatest prime number less than (or equal to !!) power(2, hash_bits)
 * * sizeof(unsigned long). Graph globals as well as hash table values
 * are initialized. */
extern GRAPH *GphCreate(/* name, hash_bits, bufsize, facts, unfold, stub,
                           delAlg, seed */);
/*
    char *name;
    unsigned hash_bits;
    unsigned bufsize;
    BOOL facts;
    BOOL unfold;
    BOOL stub;
    BOOL delAlg;
    unsigned long seed;
*/

/* GphGet returns the first node (addr.) which has not been returned by GphGet.
 * (unsigned long) -1L is returned if all nodes have already been returned.
 * Node contents are copied to buffer. */
extern unsigned long GphGet(/* graph, buffer, size */);
/*
    GRAPH *graph;
    DATATYPE *buffer;
    unsigned size;
*/

/* GphPutNode inserts a node to the node data base if not already there.
 * Return value is the data base address of the node. */
extern unsigned long GphPutNode(/* graph, node, len */);
/*
    GRAPH *graph;
    DATATYPE *node;
    unsigned len;
*/

/* GphPutArrow inserts an arrow to the arrow data base. 
 * Return value is the data base address of the arrow. */
extern unsigned long GphPutArrow(/* graph, arrow, len, ordinal, 
				    target_node_addr */);
/*
    GRAPH *graph;
    DATATYPE *arrow;
    unsigned len;
    unsigned long ordinal;
    unsigned long target_node_addr;
*/

/* GphPutFactArrow inserts a fact arrow to the arrow data base. 
 * Return value is the data base address of the arrow. */
extern unsigned long GphPutFactArrow(/* graph, arrow, len */);
/*
    GRAPH *graph;
    DATATYPE *arrow;
    unsigned len;
*/

/* GphUpdate updates the successor information and 
 * the enabled fact information of the node. */
extern void GphUpdate(/* graph, node_addr, arrow_count, first_arrow_addr,
			 fact_arrow_count, first_fact_arrow_addr */);
/*
    GRAPH *graph;
    unsigned long node_addr;
    unsigned long arrow_count;
    unsigned long first_arrow_addr;
    unsigned long fact_arrow_count;
    unsigned long first_fact_arrow_addr;
*/

/* GphAddTermNode adds a node to terminal node list. It also marks
 * the node to have no successors and updates the enabled fact information
 * of the node.
 * WARNING: It is not checked whether the node is already in the terminal node
 * list or has already been marked to have successors. */
extern void GphAddTermNode(/* graph, node_addr,
			      fact_arrow_count, first_fact_arrow_addr */);
/*
    GRAPH *graph;
    unsigned long node_addr;
    unsigned long fact_arrow_count;
    unsigned long first_fact_arrow_addr;
*/

/* GphIncGenCount increments the number of finished nodes (nodes having
 * all their successors generated) by one. */
extern void GphIncGenCount(/* graph */);
/*
    GRAPH *graph;
*/

#define GPH_GENERATE 0
#define GPH_COMPS 1
#define GPH_PROBE 2

/* GphOpen opens a graph. Corresponding files are opened.
 * Graph globals table is kept in memory when graph is open.
 * Hash table is kept in memory if graph generation is to be continued.
 * Reason is one of GPH_GENERATE, GPH_COMPS, GPH_PROBE. */
extern GRAPH *GphOpen(/* name, reason, bufsize */);
/*
    char *name;
    int reason;
    unsigned bufsize;
*/

/* GphFlush flushes the contents of graph file buffers to corresponding
 * files. In addition to this, graph header is written to file. */
extern void GphFlush(/* graph */);
/*
    GRAPH *graph;
*/

extern unsigned long GphAAG(/* graph */);
/*
    GRAPH *graph;
*/

extern unsigned long GphAAGH(/* graph */);
/*
    GRAPH *graph;
*/

#ifdef GPHQUOTA
extern void GphCOK(/* graph */);
/*
    GRAPH *graph;
*/
#endif

#ifdef UNIX
#ifdef MAPTOMEM
extern void GphMapToMem(/* graph, len */);
/*
    GRAPH *graph;
    unsigned long len;
*/
#endif
#endif

/* GphClose closes a graph. Corresponding files are closed.
 * Graph globals table and hash table (if exists) are deleted. 
 * Globals table is written to file if graph was open for construction. 
 * Hash table is written to file if graph was open for construction
 * and is not finished. */
extern void GphClose(/* graph */);
/*
    GRAPH *graph;
*/

/* GphStart returns the address of the start node. 
 * (Graph doesn't have to be finished.) */
extern unsigned long GphStart(/* graph */);
/*
    GRAPH *graph;
*/

/* GphCompNumber returns the strongly connected component 
 * number of given node (addr.). */
extern unsigned long GphCompNumber(/* graph, node_addr */);
/*
    GRAPH *graph;
    unsigned long node_addr;
*/

/* GphOrdinal returns the ordinal number of given node (addr.).
 * Start node has number 0L, etc. 
 * (Graph doesn't have to be finished.) */
extern unsigned long GphOrdinal(/* graph, node_addr */);
/*
    GRAPH *graph;
    unsigned long node_addr;
*/

/* GphId returns the address of given node (ordinal number).
 * Number should be in range 0L..(GphNdCount - 1L).
 * (Graph doesn't have to be finished.) */
extern unsigned long GphId(/* graph, ordinal */);
/*
    GRAPH *graph;
    unsigned long ordinal;
*/

/* GphNdCount returns the number of nodes of a graph.
 * Current value is always returned. (Graph doesn't have to be finished.) */
extern unsigned long GphNdCount(/* graph */);
/*
    GRAPH *graph;
*/

/* GphAwCount returns the number of arrows of given source node (addr.).
 * (unsigned long) -1L is returned if follower information is missing.
 * Any arrow number used should be in range 0L..(GphAwCount - 1L).
 * (Graph doesn't have to be finished.) */
extern unsigned long GphAwCount(/* graph, node_addr */);
/*
    GRAPH *graph;
    unsigned long node_addr;
*/

/* GphFcAwCount returns the number of fact arrows of given source node (addr.).
 * (unsigned long) -1L is returned if follower information is missing
 * and at least one fact exists.
 * Any fact arrow number used should be in range 0L..(GphFcAwCount - 1L).
 * (Graph doesn't have to be finished.) */
extern unsigned long GphFcAwCount(/* graph, node_addr */);
/*
    GRAPH *graph;
    unsigned long node_addr;
*/

/* GphFlFirst returns the first arrow of given source node.
 * (Graph doesn't have to be finished.) */
extern void GphFlFirst(/* graph, node_addr, awaddr, succ */);
/*
    GRAPH *graph;
    unsigned long node_addr;
    unsigned long *awaddr;
    unsigned long *succ;
*/

/* GphFcAwFirst returns the first fact arrow of given source node.
 * (Graph doesn't have to be finished.) */
extern void GphFcAwFirst(/* graph, node_addr, awaddr */);
/*
    GRAPH *graph;
    unsigned long node_addr;
    unsigned long *awaddr;
*/

/* GphFlNext returns the arrow next to the given arrow.
 * (Graph doesn't have to be finished.) 
 * WARNING: Caller must keep count of source nodes and arrow ordinal 
 * numbers. */
extern void GphFlNext(/* graph, awaddr, succ */);
/*
    GRAPH *graph;
    unsigned long *awaddr;
    unsigned long *succ;
*/

/* GphFcAwNext returns the fact arrow next to the given fact arrow.
 * (Graph doesn't have to be finished.) 
 * WARNING: Caller must keep count of source nodes and fact arrow ordinal 
 * numbers. */
extern void GphFcAwNext(/* graph, awaddr */);
/*
    GRAPH *graph;
    unsigned long *awaddr;
*/

/* GphNode returns the contents of given node.
 * (Graph doesn't have to be finished.)
 * WARNING: The buffer returned should neither be deallocated
 * nor modified. */
extern DATATYPE *GphNode(/* graph, node_addr */);
/*
    GRAPH *graph;
    unsigned long node_addr;
*/

/* GphArrow returns the contents of given arrow.
 * (Graph doesn't have to be finished.)
 * WARNING: The buffer returned should neither be deallocated
 * nor modified. */
extern DATATYPE *GphArrow(/* graph, awaddr */);
/*
    GRAPH *graph;
    unsigned long awaddr;
*/

/* GphFcArrow returns the contents of given fact arrow.
 * (Graph doesn't have to be finished.)
 * WARNING: The buffer returned should neither be deallocated
 * nor modified. */
extern DATATYPE *GphFcArrow(/* graph, awaddr */);
/*
    GRAPH *graph;
    unsigned long awaddr;
*/

/* GphAwOrdinal returns the ordinal number of given arrow.
 * (Graph doesn't have to be finished.) */
extern unsigned long GphAwOrdinal(/* graph, awaddr */);
/*
    GRAPH *graph;
    unsigned long awaddr;
*/

/* GphStatistics outputs the statistics of a graph. */
extern void GphStatistics(/* graph, file */);
/*
    GRAPH *graph;
    FILE *file;
*/

/* GphSgCount returns the number of strongly connected components
 * of a graph. 0L is returned if they have not been computed.
 * Any component number used should be in range 0L..(GphSgCount - 1L). */
extern unsigned long GphSgCount(/* graph */);
/*
    GRAPH *graph;
*/

/* GphSgSearch searches the strongly connected components of a graph.
 * Graph has to be finished. In other words, followers must have been
 * generated for all nodes and there must be at least one node. */
extern void GphSgSearch(/* graph */);
/*
    GRAPH *graph;
*/

/* GphSgGet returns the nodes of given strongly connected component (number).
 * Tree contains node addresses (of type unsigned long). */
extern BALTREE *GphSgGet(/* graph, component */);
/*
    GRAPH *graph;
    unsigned long component;
*/

/* GphSgFlCount returns the number of successor components
 * of given strongly connected component (number). */
extern unsigned long GphSgFlCount(/* graph, component */);
/*
    GRAPH *graph;
    unsigned long component;
*/

/* GphSgSucc returns the successor components
 * of given strongly connected component (number).
 * Tree contains component numbers (of type unsigned long). */
extern BALTREE *GphSgSucc(/* graph, component */);
/*
    GRAPH *graph;
    unsigned long component;
*/

/* GphTNCount returns the (current) number of terminal nodes of a graph.
 * (Graph doesn't have to be finished.) */
extern unsigned long GphTNCount(/* graph */);
/*
    GRAPH *graph;
*/

/* GphTNGet returns the terminal nodes of a graph.
 * Tree contains node addresses (of type unsigned long).
 * Graph doesn't have to be finished. Current terminal nodes are returned. */
extern BALTREE *GphTNGet(/* graph */);
/*
    GRAPH *graph;
*/

/* GphFcECount returns the (current) number of fact arrow source nodes
 * of a graph. (Graph doesn't have to be finished.) */
extern unsigned long GphFcECount(/* graph */);
/*
    GRAPH *graph;
*/

/* GphFcEGet returns the fact arrow source nodes of a graph.
 * Tree contains node addresses (of type unsigned long).
 * Graph doesn't have to be finished. Current fact arrow source nodes
 * are returned. */
extern BALTREE *GphFcEGet(/* graph */);
/*
    GRAPH *graph;
*/

/* GphTmSgCount returns the number of nontrivial terminal strongly connected
 * components of a graph. (A component g is terminal if and only if
 * g has no successor component. A terminal component g is trivial 
 * if and only if g is a terminal node or the graph itself.)
 * 0L is returned if strongly connected components have not
 * been computed. */
extern unsigned long GphTmSgCount(/* graph */);
/*
    GRAPH *graph;
*/

/* GphTmSgGet returns the nontrivial terminal strongly connected components
 * of a graph. Tree contains component numbers (of type unsigned long). */
extern BALTREE *GphTmSgGet(/* graph */);
/*
    GRAPH *graph;
*/

/* GphBufSize returns the size of the string buffer used during
 * graph generation. */
extern unsigned GphBufSize(/* graph */);
/*
    GRAPH *graph;
*/

/* GphBits returns the number of hash table bits of the graph. */
extern unsigned GphBits(/* graph */);
/*
    GRAPH *graph;
*/

/* GphUnfold returns TRUE iff unfolding preceded the graph generation. */
extern BOOL GphUnfold(/* graph */);
/*
    GRAPH *graph;
*/

/* GphStub returns TRUE iff stubborn set method is applied. */
extern BOOL GphStub(/* graph */);
/*
    GRAPH *graph;
*/

/* GphDelAlg returns TRUE iff stubborn set method and deletion algorithm
 * are applied. */
extern BOOL GphDelAlg(/* graph */);
/*
    GRAPH *graph;
*/

extern unsigned long GphSeed(/* graph */);
/*
    GRAPH *graph;
*/

extern void GphSetSeed(/* graph, seed */);
/*
    GRAPH *graph;
    unsigned long seed;
*/
